__package__    = "pocketphone/__init__.py"
__version__    = "1.0"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/pocketphone"
__date__       = "$Date: 2008/11/02 05:50:56 $"
__copyright__  = "Copyright (c) 2007-2008 Aaron Straup Cope. BSD license : http://www.modestmaps.com/license.txt"

from pocketindex import pocketindex

import Image
import cairo
import math
import array
import md5
        
from reportlab.pdfgen import canvas

class pocketphone :

    def __init__ (self, paper_width, paper_height, paper_margin, paper_dpi=144) :

        self.paper_width = paper_width
        self.paper_height = paper_height
        self.paper_margin = paper_margin
        self.paper_dpi = paper_dpi       

        # note the flip
        
        self.width = int(self.paper_height * self.paper_dpi)
        self.height = int(self.paper_width * self.paper_dpi)        
        self.margin = int(self.paper_margin * self.paper_dpi)

        self.sect_width = int((self.width - (self.margin * 4)) / 4)
        self.sect_height = int((self.height - (self.margin * 2)) / 2)

        self.page_map = {
            '1':'c',
            '2':'e',
            '3':'g',
            '4':'h',
            '5':'f',
            '6':'d',
            '7':'b',
            '8':'a',
        }

        self.sect_map = {
            'a' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 0},
            'b' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 0},
            
            'c' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : self.sect_width},
            'd' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : self.sect_width}, 
            
            'e' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 2 * self.sect_width},
            'f' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 2 * self.sect_width},
            
            'g' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 3 * self.sect_width},
            'h' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 3 * self.sect_width}
            }

        self.count_pages = 0

        self.queue = {}
        
        self.numbers = []
        self.pages = []
        self.sheets = []

        self.pages_map = {}
        
        self.dedupe = True
        self.sheets_rendered = []
        
    # #####################################################
    
    def add_numbers (self, numbers) :
        for data in numbers :
            self.add_number(data)
            
    # #####################################################
                
    def add_number (self, ph_data) :
        label = ph_data['name']
        number = self.format_number(ph_data['number'])

        # overkill, maybe but whatever...

        if self.dedupe :
            key = label.encode('ascii', 'replace') + number.encode('ascii', 'replace')
            hex = md5.md5(key).hexdigest()
        
            if self.queue.has_key(hex) :
                return

            self.queue[hex] = number

        data = {'label':label, 'value':number}
        self.numbers.append(data)

    #
    
    def format_number (self, number) :

        for char in ("-", "/", ".", "+", "(", ")", " ") :
            number = number.replace(char, "")

        last_four = number[-4:]
        number = number[:-4]

        last_three = number[-3:]
        number = number[:-3]

        first_three = number[-3:]
        number = number[:-3]

        prefix = number

        parts = []

        for part in (prefix, first_three, last_three, last_four) :

            clean = part.strip()
            
            if clean != '' :
                parts.append(clean)

        return " ".join(parts)
    
    # #####################################################

    def draw (self) :

        idx = pocketindex(self.sect_width, self.sect_height)
        pages = idx.create(self.numbers)

        # first and last pages are blank
        
        pages.insert(0, None)
        pages.append(None)    

        #
        
        self.count_pages = len(pages)
        self.pages = pages

        self.sheets = int(math.ceil(float(self.count_pages) / float(8)))

        # now do the layout

        first = 1
        last = (self.sheets * 8) + 1
        offset = 4

        for current_sheet in range(1, (self.sheets + 1)) :
            
            pm = Image.new('RGBA', (self.height, self.width), 'white')

            for i in range(first, (first + offset)) :
                self.draw_page(pm, i)
        
            for i in range((last - offset), last) :
                self.draw_page(pm, i)

            self.sheets_rendered.append(pm)
    
            first += offset
            last -= offset
        
    # #####################################################

    def draw_page (self, pm, pg_number) :

        if len(self.pages) < pg_number :
            return

        idx = pg_number - 1

        if not self.pages[idx] :
            return
        
        label_idx = str(self.idx_for_page_number(pg_number))
        
        sect = self.page_map[label_idx]
        meta = self.sect_map[sect]
        
        img = self.pages[ idx ]
        img = self.draw_pg_number(img, pg_number)
        
        img = img.transpose(meta['rotate'])

        # print "draw %s: %s" % (pg_number, type(img))

        x = int(meta['x'])
        y = int(meta['y'])

        x += self.margin
        y += self.margin

        if sect in ('e', 'f', 'g', 'h') :
            y += (self.margin * 2)
        
        pm.paste(img, (x,y))

    # #####################################################
    
    def draw_pg_number (self, im, pg_number) :

        (w, h) = im.size

        if pg_number % 2 :
            # odd
            x = w - 30
        else :
            # even
            x = 30
        
        y = h - 30
        
        pm = self.pil2cairo(im)
    
        ctx = cairo.Context(pm)        
        ctx.move_to(x, y)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(55, 55, 55)
        ctx.fill()
        
        ctx = cairo.Context(pm)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(0, 0, 0)
        ctx.set_line_width(2)
        ctx.stroke()
        
        ctx = cairo.Context(pm)        
        ctx.set_source_rgb(0, 0, 0)    
        
	offset_x, offset_y = self.offsets_for_page_id(ctx, pg_number)
                
        x = x + offset_x
        y = y - offset_y

        ctx.move_to(x, y)
        ctx.show_text(unicode(pg_number))

        return self.cairo2pil(pm)
        
    # #####################################################
    
    def save (self, out) :

        # reportlab docs say out can be any IO
        # thingy or a path

        # note the part where we (re) flip x, y
        # and height, width; see notes in setup()
        
        pdf = canvas.Canvas(out, pagesize=(self.height, self.width))

        for im in self.sheets_rendered :    
            pdf.drawInlineImage(im, 0, 0)
            pdf.showPage()
    
        pdf.save()

    # #####################################################

    # please put this shit in a 'pocketmod' class

    def idx_for_page_number(self, pg) :
    
        while pg > 8 :
            pg -= 8

        return pg

    # ##########################################################        
    
    def pil2cairo(self, im) :
        
        (w,h) = im.size
    
	mode = cairo.FORMAT_ARGB32

        data = im.tostring()
        a = array.array('B', data)

        return cairo.ImageSurface.create_for_data (a, mode, w, h, (w * 4))

    # ##########################################################
    
    def cairo2pil(self, surface) :

	mode='RGBA'
        
        width = surface.get_width()
        height = surface.get_height()
    
        return Image.frombuffer(mode, (width, height), surface.get_data(), "raw", mode, 0, 1)

    # #####################################################

    def offsets_for_page_id(self, ctx, pg_number) :

        # http://www.tortall.net/mu/wiki/CairoTutorial#tips-and-tricks
    
        fascent, fdescent, fheight, fxadvance, fyadvance = ctx.font_extents()
        
        xbearing, ybearing, width, height, xadvance, yadvance = (
            ctx.text_extents(unicode(pg_number)))
        
        x = 0.5 - xbearing - width / 2
        y = fdescent + fheight / 2

        # wtf do I need to adjust these again by hand?
        
    	return ((x - 1), (y - 10))

    """

    # for generating mock/example pocketphone books
    
    def mk_names(self) :
        foo = os.popen("egrep -i \"^[^aeiou']{4}\" /usr/share/dict/words")
        self.bar = map(lambda s: s.strip(), foo.readlines())
        self.barc = len(self.bar)
        
    def mk_number(self) :
        number = ''
        for i in range(0, 10):
            number += str(int(random.random() * 9))
        return self.format_number(number)
    
    def mk_label(self) :
        first = int(random.random() * self.barc)
        last = int(random.random() * self.barc)        
        return "%s %s" % (self.bar[first].title(), self.bar[last].title())

    """

if __name__ == '__main__' :

    ph = pocketphone(height, width, margin)
    ph.add_numbers(numbers)
    ph.draw(out)
